/*
 * SignPDF.cpp
 *
 *  Created on: 28.04.2017
 *      Author: huj
 */

#include <cert.h>
#include <ssl.h>

#include "SignPDF.h"

#include <poppler-form.h>
#include <poppler-qt5.h>

#include <QtCore/QDateTime>
#include <QtCore/QFile>
#include <QtCore/QTextCodec>
#include <QtCore/QTextStream>

#include <getopt.h>
#include <unistd.h>

extern char* optarg;

void help( bool err )
{
  QTextStream out( err ? stderr : stdout );
  out.setCodec( QTextCodec::codecForName( "UTF-8" ) );
  out << "Usage:\tsignPDF <options>" << endl;
  out << "Options:" << endl;
  out << "\t--in\t-i <file>\tPDF-file to be signed" << endl;
  out << "\t--out\t-o <file>\tOutput filename of signed PDF" << endl;
  out << "\t--name\t-n <nickname>\tNickname of the signing certificate" << endl;
  out << "\t--pw\t-p <password>\tPassword to unlock key database" << endl;
  out << "\t--help\t-h\t\tPrint this help message and exit" << endl;
}

int main( int argc, char** argv )
{
  int c = 0;
  int ret = 1;
  QString input;
  QString output;
  QString nickname;
  QString password;
  QTextStream err( stderr );
  err.setCodec( QTextCodec::codecForName( "UTF-8" ) );
  const char* opt_str = "hi:o:n:p:";
  struct option opt_list[] =
  {
    { "help", 0, nullptr, 'h' },
    { "in", 1, nullptr, 'i' },
    { "out", 1, nullptr, 'o' },
    { "name", 1, nullptr, 'n' },
    { "pw", 1, nullptr, 'p' },
    { nullptr, 0, nullptr, 0 }
  };
  while( ( c = getopt_long( argc, argv, opt_str, opt_list, nullptr ) ) != -1 )
  {
    switch( c )
    {
    case 'i':
      input = QString( optarg );
      break;
    case 'o':
      output = QString( optarg );
      break;
    case 'n':
      nickname = QString( optarg );
      break;
    case 'p':
      password = QString( optarg );
      break;
    case 'h':
      help( false );
      return 0;
    case '?':
      help( true );
      return 1;
    }
  }
  if( input.isEmpty() )
  {
    err << "Error: An input PDF file must be given" << endl;
  }
  else if( output.isEmpty() || output == input )
  {
    err << "Error: An output PDF file other than input file must be given" << endl;
  }
  else
  {
    QFile inFile( input );
    if( inFile.exists() )
    {
      SignPDF signPDF( input, inFile.size() );
      ret = signPDF.sign( nickname, password, output );
    }
    else
    {
      err << "Error: PDF file '" << input << "' does not exist" << endl;
    }
  }
  return ret;
}

SignPDF::SignPDF( const QString& file, qint64 s )
  : size( s ), pdfFile( file ),
    popplerDoc( nullptr ), popplerPage( nullptr ), sigField( nullptr )
{
}

SignPDF::~SignPDF()
{
  delete sigField;
  delete popplerPage;
}

int SignPDF::sign( const QString& nickname, const QString& password, const QString& file )
{
  int ret = 0;
  QTextStream out( stdout );
  out.setCodec( QTextCodec::codecForName( "UTF-8" ) );

  popplerDoc = Poppler::Document::load( pdfFile );
  if( popplerDoc != nullptr )
  {
    if( popplerDoc->isLocked() )
    {
      out << "PDF file '" << pdfFile << "' is locked" << endl;
    }
    else
    {
      for( int n = 0; n < popplerDoc->numPages(); ++n )
      {
        popplerPage = popplerDoc->page( n );
        if( searchSignatureField() )
        {
          break;
        }
        else
        {
          delete popplerPage;
          popplerPage = nullptr;
        }
      }
      if( sigField != nullptr )
      {
        sigField->setSignatureType( Poppler::FormFieldSignature::EtsiCAdESdetached );
        if( sigField->sign( nickname, password ) )
        {
          Poppler::PDFConverter* pdfConverter = popplerDoc->pdfConverter();
          pdfConverter->setOutputFileName(file);
          pdfConverter->setPDFOptions(pdfConverter->pdfOptions() |
                                      Poppler::PDFConverter::WithChanges);
          pdfConverter->convert();
          delete pdfConverter;
        }
        else
        {
          out << "Could not sign '" << pdfFile << "'" << endl;
        }
      }
    }
    delete popplerDoc;
  }
  else
  {
    out << "Could not load '" << pdfFile << "'" << endl;
  }

  return ret;
}

bool SignPDF::searchSignatureField()
{
  for( Poppler::FormField* field : popplerPage->formFields() )
  {
    if( sigField == nullptr && field->type() == Poppler::FormField::FormSignature )
    {
      sigField = dynamic_cast<Poppler::FormFieldSignature*>( field );
    }
    else
    {
      delete field;
    }
  }
  return sigField != nullptr;
}
